/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 *
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Header: /nfs/jade/vint/CVSROOT/ns-2/mac-802_11.cc,v 1.38 2000/11/21 18:58:47 johnh Exp $
 *
 * Ported from CMU/Monarch's code, nov'98 -Padma.
 */

#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"
#include "mobilenode.h"

// #define DEBUG

#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "mac-timers.h"
#include "mac-802_11.h"
#include "cmu-trace.h"


// XXX Can't we make these macros inline methods? Otherwise why should we have
// inline methods at all??

//dscr
#define CHECK_BACKOFF_TIMER2()                                           \
{                                                                       \
        if(is_idle() && mhBackoff2_.paused()) {                          \
                mhBackoff2_.resume(difs_);                             \
        }                                                               \
        if(!is_idle() && mhBackoff2_.busy() && !mhBackoff2_.paused()){   \
                mhBackoff2_.pause();                                     \
        }                                                               \
}
//dscr
//dscr
#define CHECK_BACKOFF_TIMER()						\
{									\
	if(is_idle() && mhBackoff_.paused() && (recv_ == 0)) {		\
		mhBackoff_.resume(FR_Time);				\
	}								\
	if(!is_idle() && mhBackoff_.busy() && !mhBackoff_.paused()){	\
		mhBackoff_.pause();					\
	}								\
}
//dscr

#define TRANSMIT(p, t)                                                  \
{       								\
	tx_active_ = 1;                                                  \
                                                                        \
        /*                                                              \
         * If I'm transmitting without doing CS, such as when           \
         * sending an ACK, any incoming packet will be "missed"         \
         * and hence, must be discarded.                                \
         */                                                             \
        if(rx_state_ != MAC_IDLE) {                                      \
                struct hdr_mac802_11 *dh = HDR_MAC802_11(p);                  \
                                                                        \
                assert(dh->dh_fc.fc_type == MAC_Type_Control);          \
                assert(dh->dh_fc.fc_subtype == MAC_Subtype_ACK);        \
                                                                        \
                assert(pktRx_);                                          \
                struct hdr_cmn *ch = HDR_CMN(pktRx_);                    \
                                                                        \
                ch->error() = 1;        /* force packet discard */      \
        }                                                               \
                                                                        \
        /*                                                              \
         * pass the packet on the "interface" which will in turn        \
         * place the packet on the channel.                             \
         *                                                              \
         * NOTE: a handler is passed along so that the Network          \
         *       Interface can distinguish between incoming and         \
         *       outgoing packets.                                      \
         */                                                             \
        downtarget_->recv(p->copy(), this);                             \
                                                                        \
        mhSend_.start(t);                                                \
                                                                        \
	mhIF_.start(TX_Time(p));                                         \
}
//dscr
#define SET_RX_STATE(x)			\
{					\
	rx_state_ = (x);		\
	CHECK_BACKOFF_TIMER();		\
        CHECK_BACKOFF_TIMER2();          \
}

#define SET_TX_STATE(x)				\
{						\
	tx_state_ = (x);			\
	CHECK_BACKOFF_TIMER();			\
        CHECK_BACKOFF_TIMER2();          \
}
//dscr

/* ======================================================================
   Global Variables
   ====================================================================== */
static PHY_MIB PMIB =
{
//dscr
	FS_CWMin, FS_CWMax,
//dscr
	DSSS_CWMin, DSSS_CWMax, DSSS_SlotTime, DSSS_CCATime,
	DSSS_RxTxTurnaroundTime, DSSS_SIFSTime, DSSS_PreambleLength,
	DSSS_PLCPHeaderLength
};	

#ifdef RTS_HANDSHAKE
static MAC_MIB MMIB =
{
        0 /* MAC_RTSThreshold */, MAC_ShortRetryLimit,
        MAC_LongRetryLimit, MAC_FragmentationThreshold,
        MAC_MaxTransmitMSDULifetime, MAC_MaxReceiveLifetime,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};      
#else
static MAC_MIB MMIB =
{
        3000 /* MAC_RTSThreshold */, MAC_ShortRetryLimit,
        MAC_LongRetryLimit, MAC_FragmentationThreshold,
        MAC_MaxTransmitMSDULifetime, MAC_MaxReceiveLifetime,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
#endif

/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */
static class Mac802_11Class : public TclClass {
public:
	Mac802_11Class() : TclClass("Mac/802_11") {}
	TclObject* create(int, const char*const*) {
		return (new Mac802_11(&PMIB, &MMIB));
	}
} class_mac802_11;


/* ======================================================================
   Mac Class Functions
   ====================================================================== */
Mac802_11::Mac802_11(PHY_MIB *p, MAC_MIB *m) : Mac(),
	mhIF_(this), mhNav_(this), mhRecv_(this), mhSend_(this),
	mhDefer_(this, p->SlotTime), mhBackoff_(this, p->SlotTime), mhBackoff2_(this, p->SlotTime)  //dscr
{
#ifdef YX_DELAY
        bind("StartTime_", &StartTime_);
#endif

	macmib_ = m;
	phymib_ = p;

	nav_ = 0.0;

	tx_state_ = rx_state_ = MAC_IDLE;
	tx_active_ = 0;

	pktRTS_ = 0;
	pktCTRL_ = 0;
//dscr
        bind("FS_K", &FS_K);
        bind("CWMin_2nd", &CWMin_2nd);

        phymib_->CWMin = CWMin_2nd;
 
	scw_ = phymib_->CWMin;
	fcw_ = phymib_->F_CWMin;
	K = FS_K;
	phase = 1;
	threshTime = 0.0;
	recv_ = 0;
	nav_category = 1;
//dscr
	ssrc_ = slrc_ = 0;

	sifs_ = phymib_->SIFSTime;
	pifs_ = sifs_ + phymib_->SlotTime;
	difs_ = sifs_ + 2*phymib_->SlotTime;
//xueyang, the physical layer preamble has been counted twice.
	eifs_ = sifs_ + difs_ + DATA_Time(ETHER_ACK_LEN);

	tx_sifs_ = sifs_ - phymib_->RxTxTurnaroundTime;
	tx_pifs_ = tx_sifs_ + phymib_->SlotTime;
	tx_difs_ = tx_sifs_ + 2 * phymib_->SlotTime;

	sta_seqno_ = 1;
	cache_ = 0;
	cache_node_count_ = 0;
}
	

int
Mac802_11::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "log-target") == 0) {
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		} else if(strcmp(argv[1], "nodes") == 0) {
			if(cache_) return TCL_ERROR;
			cache_node_count_ = atoi(argv[2]);
			cache_ = new Host[cache_node_count_ + 1];
			assert(cache_);
			bzero(cache_, sizeof(Host) * (cache_node_count_+1 ));
			return TCL_OK;
		}
	}
	return Mac::command(argc, argv);
}

/* ======================================================================
   Debugging Routines
   ====================================================================== */
void
Mac802_11::trace_pkt(Packet *p) {
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		ETHER_ADDR(dh->dh_da), ETHER_ADDR(dh->dh_sa),
		index_, packet_info.name(ch->ptype()), ch->size());
}

void
Mac802_11::dump(char *fname)
{
	fprintf(stderr,
		"\n%s --- (INDEX: %d, time: %2.9f)\n",
		fname, index_, Scheduler::instance().clock());

	fprintf(stderr,
		"\ttx_state_: %x, rx_state_: %x, nav: %2.9f, idle: %d\n",
		tx_state_, rx_state_, nav_, is_idle());

	fprintf(stderr,
		"\tpktTx_: %x, pktRx_: %x, pktRTS_: %x, pktCTRL_: %x, callback: %x\n",
		(int) pktTx_, (int) pktRx_, (int) pktRTS_,
		(int) pktCTRL_, (int) callback_);

	fprintf(stderr,
		"\tDefer: %d, Backoff: %d (%d), Recv: %d, Timer: %d Nav: %d\n",
		mhDefer_.busy(), mhBackoff_.busy(), mhBackoff_.paused(),
		mhRecv_.busy(), mhSend_.busy(), mhNav_.busy());
	fprintf(stderr,
		"\tBackoff Expire: %f\n",
		mhBackoff_.expire());
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */
inline int
Mac802_11::hdr_dst(char* hdr, int dst )
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
	//dst = (u_int32_t)(dst);

	if(dst > -2)
		STORE4BYTE(&dst, (dh->dh_da));

	return ETHER_ADDR(dh->dh_da);
}

inline int 
Mac802_11::hdr_src(char* hdr, int src )
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
	if(src > -2)
		STORE4BYTE(&src, (dh->dh_sa));
	return ETHER_ADDR(dh->dh_sa);
}

inline int 
Mac802_11::hdr_type(char* hdr, u_int16_t type)
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
	if(type)
		STORE2BYTE(&type,(dh->dh_body));
	return GET2BYTE(dh->dh_body);
}


/* ======================================================================
   Misc Routines
   ====================================================================== */
inline int
Mac802_11::is_idle()
{
	if(rx_state_ != MAC_IDLE){
		//printf("(%d) rx_state_ is not idle:%d\n", index_, rx_state_);
		return 0;
	}
	if(tx_state_ != MAC_IDLE){
                //printf("(%d) tx_state_ is not idle:%d\n", index_, tx_state_);
		return 0;
	}
	if(nav_ > Scheduler::instance().clock()){
                //printf("(%d) nav is not idle\n", index_);
		return 0;
	}
        //printf("(%d) is idle!!\n", index_);
	return 1;
}

void
Mac802_11::discard(Packet *p, const char* why)
{
	hdr_mac802_11* mh = HDR_MAC802_11(p);
	hdr_cmn *ch = HDR_CMN(p);

#if 0
	/* old logic 8/8/98 -dam */
	/*
	 * If received below the RXThreshold, then just free.
	 */
	if(p->txinfo_.Pr < p->txinfo_.ant.RXThresh) {
		Packet::free(p);
		//p = 0;
		return;
	}
#endif // 0

	/* if the rcvd pkt contains errors, a real MAC layer couldn't
	   necessarily read any data from it, so we just toss it now */
	if(ch->error() != 0) {
		Packet::free(p);
		//p = 0;
		return;
	}

	switch(mh->dh_fc.fc_type) {
	case MAC_Type_Management:
		drop(p, why);
		return;
	case MAC_Type_Control:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_RTS:
			if((u_int32_t)ETHER_ADDR(mh->dh_sa) == \
			   (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			/* fall through - if necessary */
		case MAC_Subtype_CTS:
		case MAC_Subtype_ACK:
			if((u_int32_t)ETHER_ADDR(mh->dh_da) == \
			   (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Control subtype\n");
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_Data:
			if((u_int32_t)ETHER_ADDR(mh->dh_da) == \
			   (u_int32_t)index_ ||
			   (u_int32_t)ETHER_ADDR(mh->dh_sa) == \
			   (u_int32_t)index_ ||
			   (u_int32_t)ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
				drop(p);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Data subtype\n");
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "invalid MAC type (%x)\n", mh->dh_fc.fc_type);
		trace_pkt(p);
		exit(1);
	}
	Packet::free(p);
}

void
Mac802_11::capture(Packet *p)
{
	/*
	 * Update the NAV so that this does not screw
	 * up carrier sense.
	 */
	//printf("(%d) set_nav at capturing\n", index_);
	set_nav(usec(eifs_ + TX_Time(p)));
	Packet::free(p);
}

void
Mac802_11::collision(Packet *p)
{
	switch(rx_state_){
	case MAC_RECV:
		SET_RX_STATE(MAC_COLL);
		/* fall through */
	case MAC_COLL:
		assert(pktRx_);
		assert(mhRecv_.busy());
		/*
		 *  Since a collision has occurred, figure out
		 *  which packet that caused the collision will
		 *  "last" the longest.  Make this packet,
		 *  pktRx_ and reset the Recv Timer if necessary.
		 */
		if(TX_Time(p) > mhRecv_.expire()) {
			mhRecv_.stop();
			discard(pktRx_, DROP_MAC_COLLISION);
			pktRx_ = p;
			mhRecv_.start(TX_Time(pktRx_));
		}
		else {
			discard(p, DROP_MAC_COLLISION);
		}
		break;
	default:
		assert(0);
	}
}

void
Mac802_11::tx_resume()
{
	assert(mhSend_.busy() == 0);
	//fprintf(stderr, "node:%d\n",index_);
	assert(mhDefer_.busy() == 0);

	if(pktCTRL_) {
		/*
		 *  Need to send a CTS or ACK.
		 */
		//printf("(%d)start defer timer at tx_resume() for pktCTRL_\n", index_);
		mhDefer_.start(sifs_);
	} else if(pktRTS_) {
//dscr
		if((mhBackoff_.busy() == 0) && (mhBackoff2_.busy() == 0)){
			mhDefer_.start(difs_);
	                //printf("(%d)start defer timer at tx_resume() for pktRTS_\n", index_);
		}
	} else if(pktTx_) {
//dscr
		//printf("(%u) in tx_resume(), phase is :%u\n", index_, phase);
                if((mhBackoff_.busy() == 0) && (mhBackoff2_.busy() == 0)){
 	                //printf("(%d)start defer timer at tx_resume() for pktTx_\n", index_);
			mhDefer_.start(sifs_);    //change from difs_ to sifs_ *******
		}
	} else if(callback_) {
		Handler *h = callback_;
		callback_ = 0;
		h->handle((Event*) 0);
	}
	SET_TX_STATE(MAC_IDLE);
}

void
Mac802_11::rx_resume()
{
	assert(pktRx_ == 0);
	assert(mhRecv_.busy() == 0);
	SET_RX_STATE(MAC_IDLE);
}


/* ======================================================================
   Timer Handler Routines
   ====================================================================== */
void
Mac802_11::backoffHandler1()
{
	//printf("(%d) calling backoffHandler1\n", index_);
/*dscr
	if(pktCTRL_) {
		//printf("(%u) hi_1\n", index_);
		assert(mhSend_.busy() || mhDefer_.busy());
		return;
	}

	if(phase == 1) {
           //printf("(%u) In backoffHandler(), phase is changed from 1 to 2\n", index_);
	   phase = 2;
	}
        if((phase == 2) &&(pktRTS_ || pktTx_ )){
            //printf("(%u) hi_3, mhBackoff2 is started\n", index_);
	    assert(mhBackoff2_.busy() == 0);
            mhBackoff2_.start(scw_, is_idle());
	    if(!mhBackoff2_.paused()){
		mhBackoff2_.pause();
		mhBackoff2_.resume(difs_);
	    }
        }
dscr*/
//dscr  when directly transmit from the first stage, we still have difs_ to wait
        if((phase == 1) && (pktRTS_ || pktTx_)) {
           //printf("(%u) In backoffHandler1(), phase is changed from 1 to 2\n", index_);
           phase = 2;
	   threshTime = 0.0;   //nonlinear decrement
           Tcl::instance().evalf("colincrease");
	   Tcl::instance().evalf("interferenceIncrease");
        }

	if((phase == 2) && (pktRTS_ || pktTx_)) { 
	   if (recv_ == 1){   // DeferTimer will be busy when it is in SIFS after received RTS, Data 
                phase = 1;
                //printf("(%u) As receiver, phase changes from 2 to 1 at backoffHandler1()\n", index_);
                inc_cw();
                //printf("(%u) In backoffHandler1, fcw_ is doubled: %u\n", index_, fcw_);
		mhBackoff_.start(fcw_, is_idle());
	   } else {    
	   	assert(mhDefer_.busy() == 0);
	   	//printf("(%u) start DeferTimer in backoffHandler1()\n", index_);
	   	mhDefer_.start(difs_);
	   }
	}
//dscr
}
//dscr
void
Mac802_11::backoffHandler()
{   
        if(pktCTRL_) {  
                assert(mhSend_.busy() || mhDefer_.busy());
                return;
        }
                
        if(check_pktRTS() == 0)
                return; 
 
        if(check_pktTx() == 0)
                return;
}      
//dscr
void
Mac802_11::deferHandler()
{
	assert(pktCTRL_ || pktRTS_ || pktTx_);

        //printf("(%d)deferHandler!\n", index_);

	if(check_pktCTRL() == 0)
		return;
	assert(mhBackoff_.busy() == 0);
//dscr
	assert(mhBackoff2_.busy() == 0);
//dscr
	//if (mhBackoff_.busy() != 0)
	//{
	//	printf("deferHandler:mhBackoff_ busy!\n");
	//	return;
	//}
	if(check_pktRTS() == 0)
		return;

	if(check_pktTx() == 0)
		return;
}

void
Mac802_11::navHandler()
{
//dscr
	Phase_Check();
//dscr
}
//dscr  Phase_Check() is just designed for navHandler.
void
Mac802_11::Phase_Check()
{
   double temp;
   if((phase == 1) && (pktRTS_ || pktTx_)){
     if(nav_category == 1) {
	nav_category = 0;
        if(mhBackoff_.busy() && !mhBackoff_.paused()) {
           mhBackoff_.pause();   //update FR_Time in mhBackoff_.pause()
        }
	temp = K * phymib_->SlotTime;
	if(threshTime == 0.0){
           threshTime = K * phymib_->SlotTime;   //nonlinear decrement
	   phase1_count = 0;
	   threshCount = 1;
	}
/*	else if (phase1_count <= 2){
           threshTime = K * phymib_->SlotTime;   //nonlinear decrement
           phase1_count ++ ;
           threshCount = 2;
	} */ 
	else {
	   threshTime = K * phymib_->SlotTime * threshCount;
	   threshCount = threshCount * 2;
	} 
//        threshTime = K * phymib_->SlotTime;   //linear decrement  
 
        if (FR_Time > threshTime) {
           FR_Time -= threshTime;
           //printf("%u in navHandler(), FR_Time now is :%lf, threshTime now is: %lf\n", 
	   //		index_, FR_Time, threshTime);
           if(is_idle() && mhBackoff_.paused())
                mhBackoff_.resume(FR_Time);
        } else {
           FR_Time = 0.0;
           if(mhBackoff_.busy() || mhBackoff_.paused()){
                //printf("(%u) stop backoff timer at position s1\n", index_);
                mhBackoff_.stop();
           }
           phase = 2;
	   threshTime = 0.0;     //nonlinear decrement
           //printf("(%u) phase is changed from 1 to 2 in Phase_Check(). \n", index_);
           cwtime = Scheduler::instance().clock();
           //printf("%d %10.9g 3 %d\n", index_,cwtime,scw_);
           Tcl::instance().evalf("colincrease");
        }
     } else {
	if(is_idle() && mhBackoff_.paused())
	   mhBackoff_.resume(FR_Time);
     }
   } 
   if((phase == 2) && ((pktRTS_) || (pktTx_))){
	if(mhBackoff2_.busy() == 0){   
//there is might be nav  setting and expire for collisions.
	   //printf("(%u) mhBackoff2 is started in phase_check with scw: %d !\n", index_, scw_);
           mhBackoff2_.start(scw_, is_idle());
            if(!mhBackoff2_.paused()){    //considering the difs intermission
                mhBackoff2_.pause();
                mhBackoff2_.resume(difs_);
            }
	}
	if(is_idle() && mhBackoff2_.paused())
           mhBackoff2_.resume(difs_);

   }
}
//dscr
void
Mac802_11::recvHandler()
{
	recv_timer();
}

void
Mac802_11::sendHandler()
{
	send_timer();
}


void
Mac802_11::txHandler()
{
	tx_active_ = 0;
}


/* ======================================================================
   The "real" Timer Handler Routines
   ====================================================================== */
void
Mac802_11::send_timer()
{
	switch(tx_state_) {
	/*
	 * Sent a RTS, but did not receive a CTS.
	 */
	case MAC_RTS:
		//printf("(%d) send_timer, retransmitRTS\n", index_);
		RetransmitRTS();
		break;
	/*
	 * Sent a CTS, but did not receive a DATA packet.
	 */
	case MAC_CTS:
		assert(pktCTRL_);
		Packet::free(pktCTRL_); pktCTRL_ = 0;
		break;
	/*
	 * Sent DATA, but did not receive an ACK packet.
	 */
	case MAC_SEND:
		RetransmitDATA();
		break;
	/*
	 * Sent an ACK, and now ready to resume transmission.
	 */
	case MAC_ACK:
		assert(pktCTRL_);
		Packet::free(pktCTRL_); pktCTRL_ = 0;
//dscr
		recv_ = 0;
                if(mhBackoff_.busy() == 0) {
		   if(mhBackoff2_.busy()) mhBackoff2_.stop();
                   //printf("(%u) start backoff timer at position b5\n", index_);
		   //printf("(%u) mhBackoff2_ busy? %u, paused? %u\n", index_, mhBackoff2_.busy(), mhBackoff2_.paused());
                   mhBackoff_.start(fcw_, is_idle());
                }
//dscr
		break;
	case MAC_IDLE:
		break;
	default:
		assert(0);
	}
	tx_resume();
}


/* ======================================================================
   Outgoing Packet Routines
   ====================================================================== */
int
Mac802_11::check_pktCTRL()
{
	struct hdr_mac802_11 *mh;
	double timeout;
	double now;

        //printf("\n (%u) come to check_pktCTRL() \n", index_);

	if(pktCTRL_ == 0)
		return -1;
	if(tx_state_ == MAC_CTS || tx_state_ == MAC_ACK)
		return -1;

	mh = HDR_MAC802_11(pktCTRL_);
							  
	switch(mh->dh_fc.fc_subtype) {
	/*
	 *  If the medium is not IDLE, don't send the CTS.
	 */
	case MAC_Subtype_CTS:
		if(!is_idle()) {
			discard(pktCTRL_, DROP_MAC_BUSY); pktCTRL_ = 0;
			return 0;
		}
		SET_TX_STATE(MAC_CTS);
		timeout = (mh->dh_duration * 1e-6) + CTS_Time; // XXX
                now = Scheduler::instance().clock();
                //printf("\n (%u) send CTS at (%10.8g)\n", index_, now);
		break;
	/*
	 * IEEE 802.11 specs, section 9.2.8
	 * Acknowledments are sent after an SIFS, without regard to
	 * the busy/idle state of the medium.
	 */
	case MAC_Subtype_ACK:
		SET_TX_STATE(MAC_ACK);
		timeout = ACK_Time;
                //printf("\n (%u) send ACK\n", index_);
		break;
	default:
		fprintf(stderr, "check_pktCTRL:Invalid MAC Control subtype\n");
		exit(1);
	}
        TRANSMIT(pktCTRL_, timeout);
	return 0;
}

int
Mac802_11::check_pktRTS()
{
	struct hdr_mac802_11 *mh;
	double timeout;
	double now;

	//printf("(%d) check_pktRTS: mhBackoff_.busy? : %d, pause?: %d\n", index_, mhBackoff_.busy(), mhBackoff_.paused());
	assert(mhBackoff_.busy() == 0);
//dscr
	assert(mhBackoff2_.busy() == 0); 
//dscr

	if(pktRTS_ == 0)
 		return -1;

        now = Scheduler::instance().clock();
//dscr
	//struct hdr_cmn *ch = HDR_CMN(pktRTS_);
	mh = HDR_MAC802_11(pktRTS_);

 	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_RTS:
		if(!is_idle()) {
//dscr
		    inc_scw();
		    //inc_cw();
		    //printf("(%u) in check_pktRTS, channel is not idle, mhBackoff2 is started\n", index_);
		    //printf("(%u) scw_ is doubled: %u\n", index_, scw_);
		    mhBackoff2_.start(scw_, is_idle());
		    return 0;
		}
//dscr
                //printf("\n(%u) Transmit RTS at (%10.8g), phase is: %u\n", index_, now, phase);

		SET_TX_STATE(MAC_RTS);
		timeout = CTSTimeout;
		break;
	default:
		fprintf(stderr, "check_pktRTS:Invalid MAC Control subtype\n");
		exit(1);
	}
        TRANSMIT(pktRTS_, timeout);
	return 0;
}

int
Mac802_11::check_pktTx()
{
	struct hdr_mac802_11 *mh;
	double timeout;
	
	assert(mhBackoff_.busy() == 0);
//dscr
	assert(mhBackoff2_.busy() == 0);
//dscr

	if(pktTx_ == 0)
		return -1;

	mh = HDR_MAC802_11(pktTx_);
       	int len = HDR_CMN(pktTx_)->size();

	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_Data:
		if(! is_idle()) {
			sendRTS(ETHER_ADDR(mh->dh_da));
//dscr
			inc_scw();
			//inc_cw();
			//printf("(%u) in check_pktTx, channel is not idle, mhBackoff2 is started\n", index_);
			//printf("(%u) scw_ is doubled: %u\n", index_, scw_);
			mhBackoff2_.start(scw_, is_idle());
                        return 0;
                }
//dscr
	        //printf("\n(%u) Transmit Data\n", index_);
		SET_TX_STATE(MAC_SEND);
		if((u_int32_t)ETHER_ADDR(mh->dh_da) != MAC_BROADCAST)
			//timeout = ACKTimeout(netif_->txtime(pktTx_))+5;
			// why 10 ? buggy
			//timeout = ACKTimeout(len) + 10;
			timeout = ACKTimeout(len);
		else
			timeout = TX_Time(pktTx_);
		break;
	default:
		fprintf(stderr, "check_pktTx:Invalid MAC Control subtype\n");
		//printf("pktRTS:%x, pktCTS/ACK:%x, pktTx:%x\n",pktRTS_, pktCTRL_,pktTx_);
		exit(1);
	}
        TRANSMIT(pktTx_, timeout);
	return 0;
}

/*
 * Low-level transmit functions that actually place the packet onto
 * the channel.
 */
void
Mac802_11::sendRTS(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);
	
	assert(pktTx_);
	assert(pktRTS_ == 0);

	/*
	 *  If the size of the packet is larger than the
	 *  RTSThreshold, then perform the RTS/CTS exchange.
	 *
	 *  XXX: also skip if destination is a broadcast
	 */
//dscr
        if((u_int32_t) dst == MAC_BROADCAST) {
		scw_ = BROADCAST_CW;
	}
//dscr

	if( (u_int32_t) HDR_CMN(pktTx_)->size() < macmib_->RTSThreshold ||
	    (u_int32_t) dst == MAC_BROADCAST) {
		Packet::free(p);
		//p = 0;
		return;
	}

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_RTS_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(rf, MAC_HDR_LEN);

	rf->rf_fc.fc_protocol_version = MAC_ProtocolVersion;
 	rf->rf_fc.fc_type	= MAC_Type_Control;
 	rf->rf_fc.fc_subtype	= MAC_Subtype_RTS;
 	rf->rf_fc.fc_to_ds	= 0;
 	rf->rf_fc.fc_from_ds	= 0;
 	rf->rf_fc.fc_more_frag	= 0;
 	rf->rf_fc.fc_retry	= 0;
 	rf->rf_fc.fc_pwr_mgt	= 0;
 	rf->rf_fc.fc_more_data	= 0;
 	rf->rf_fc.fc_wep	= 0;
 	rf->rf_fc.fc_order	= 0;

	rf->rf_duration = RTS_DURATION(pktTx_);
	//ETHER_ADDR(rf->rf_ra) = dst;
	STORE4BYTE(&dst, (rf->rf_ra));

	//ETHER_ADDR(rf->rf_ta) = index_;
	STORE4BYTE(&index_, (rf->rf_ta));

	pktRTS_ = p;
}

void
Mac802_11::sendCTS(int dst, double rts_duration)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct cts_frame *cf = (struct cts_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_CTS_LEN;
	ch->iface() = -2;
	ch->error() = 0;
	//ch->direction() = hdr_cmn::DOWN;
	bzero(cf, MAC_HDR_LEN);

	cf->cf_fc.fc_protocol_version = MAC_ProtocolVersion;
	cf->cf_fc.fc_type	= MAC_Type_Control;
	cf->cf_fc.fc_subtype	= MAC_Subtype_CTS;
 	cf->cf_fc.fc_to_ds	= 0;
 	cf->cf_fc.fc_from_ds	= 0;
 	cf->cf_fc.fc_more_frag	= 0;
 	cf->cf_fc.fc_retry	= 0;
 	cf->cf_fc.fc_pwr_mgt	= 0;
 	cf->cf_fc.fc_more_data	= 0;
 	cf->cf_fc.fc_wep	= 0;
 	cf->cf_fc.fc_order	= 0;
	
	cf->cf_duration = CTS_DURATION(rts_duration);
	//ETHER_ADDR(cf->cf_ra) = dst;
	STORE4BYTE(&dst, (cf->cf_ra));
	//STORE4BYTE(&dst, (mh->dh_da));

	pktCTRL_ = p;
	
}

void
Mac802_11::sendACK(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct ack_frame *af = (struct ack_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_ACK_LEN;
	ch->iface() = -2;
	ch->error() = 0;
	
	bzero(af, MAC_HDR_LEN);

	af->af_fc.fc_protocol_version = MAC_ProtocolVersion;
 	af->af_fc.fc_type	= MAC_Type_Control;
 	af->af_fc.fc_subtype	= MAC_Subtype_ACK;
 	af->af_fc.fc_to_ds	= 0;
 	af->af_fc.fc_from_ds	= 0;
 	af->af_fc.fc_more_frag	= 0;
 	af->af_fc.fc_retry	= 0;
 	af->af_fc.fc_pwr_mgt	= 0;
 	af->af_fc.fc_more_data	= 0;
 	af->af_fc.fc_wep	= 0;
 	af->af_fc.fc_order	= 0;

	af->af_duration = ACK_DURATION();
	//ETHER_ADDR(af->af_ra) = dst;
	STORE4BYTE(&dst, (af->af_ra));

	pktCTRL_ = p;
}

void
Mac802_11::sendDATA(Packet *p)
{
	hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);

	assert(pktTx_ == 0);

	/*
	 * Update the MAC header
	 */
	ch->size() += ETHER_HDR_LEN;

	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Data;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Data;
	//printf(".....p = %x, mac-subtype-%d\n",p,dh->dh_fc.fc_subtype);
	
	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;

	if((u_int32_t)ETHER_ADDR(dh->dh_da) != MAC_BROADCAST)
		dh->dh_duration = DATA_DURATION();
         else
		dh->dh_duration = 0;
	pktTx_ = p;
}

/* ======================================================================
   Retransmission Routines
   ====================================================================== */
void
Mac802_11::RetransmitRTS()
{
	assert(pktTx_);
	assert(pktRTS_);
	assert(mhBackoff_.busy() == 0);

	macmib_->RTSFailureCount++;
        Tcl::instance().evalf("retryincrease");

#ifdef RETRANSMIT_LIMIT
	ssrc_ += 1;			// STA Short Retry Count
#endif

	if(ssrc_ >= macmib_->ShortRetryLimit) {
		discard(pktRTS_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktRTS_ = 0;
		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        /*
                         *  Need to remove the MAC header so that 
                         *  re-cycled packets don't keep getting
                         *  bigger.
                         */
                        ch->size() -= ETHER_HDR_LEN;
                        ch->xmit_reason_ = XMIT_REASON_RTS;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }
		//printf("(%d)....discarding RTS:%x\n",index_,pktRTS_);
		discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktTx_ = 0;
		ssrc_ = 0;
//dscr
		rst_cw();
		scw_ = phymib_->CWMin;
                cwtime = Scheduler::instance().clock();
 
                //printf("%d %10.9g 2 %d\n", index_,cwtime, scw_);

        	phase = 1;
		threshTime = 0.0;  //nonlinear decrement
		//printf("(%u) phase is reset to 1 when exceeding retry limit\n", index_);
//dscr     
	} else {
		//printf("(%d)...retxing RTS:%x\n",index_,pktRTS_);
		struct rts_frame *rf;
		rf = (struct rts_frame*)pktRTS_->access(hdr_mac::offset_);
		rf->rf_fc.fc_retry = 1;

//dscr
		//inc_cw();
		inc_scw();
		//printf("(%u) In RetransmitRTS(), mhBackoff2 is started, scw_ is doubled: %u, phase is : %u\n", index_, scw_, phase);
                mhBackoff2_.start(scw_, is_idle());
//dscr
	}
}

void
Mac802_11::RetransmitDATA()
{
	struct hdr_cmn *ch;
	struct hdr_mac802_11 *mh;
	u_int32_t *rcount, *thresh;

	assert(mhBackoff_.busy() == 0);
//dscr
	assert(mhDefer_.busy() == 0);
//dscr

	assert(pktTx_);
	assert(pktRTS_ == 0);

	ch = HDR_CMN(pktTx_);
	mh = HDR_MAC802_11(pktTx_);

	/*
	 *  Broadcast packets don't get ACKed and therefore
	 *  are never retransmitted.
	 */
	if((u_int32_t)ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
		Packet::free(pktTx_); pktTx_ = 0;

		/*
		 * Backoff at end of TX.
		 */
		rst_cw();
//dscr
		//rst_scw();

                cwtime = Scheduler::instance().clock();
                 
                //printf("%d %10.9g 2 %d\n", index_,cwtime, scw_);

		phase = 1;
		threshTime = 0.0;  //nonlinear decrement
               
                //printf("(%u) After Broadcast packet, phase is set to 1, start backoff timer\n", index_);

                mhBackoff_.start(fcw_, is_idle());
		
//dscr
		return;
	}

	macmib_->ACKFailureCount++;

	if((u_int32_t) ch->size() <= macmib_->RTSThreshold) {
		rcount = &ssrc_;
		thresh = &macmib_->ShortRetryLimit;
	}
	else {
		rcount = &slrc_;
		thresh = &macmib_->LongRetryLimit;
	}

	(*rcount)++;

	if(*rcount > *thresh) {
		macmib_->FailedCount++;
		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        ch->size() -= ETHER_HDR_LEN;
                        ch->xmit_reason_ = XMIT_REASON_ACK;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }

		discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktTx_ = 0;
		//printf("(%d)DATA discarded: count exceeded\n",index_);
		*rcount = 0;
//dscr
                rst_cw();
		scw_ = phymib_->CWMin;
                cwtime = Scheduler::instance().clock();
         
                //printf("%d %10.9g 2 %d\n", index_,cwtime, scw_);

                phase = 1;
		threshTime = 0.0; 	//nonlinear decrement
                //printf("(%u) phase is reset to 1 when exceeding retry limit\n", index_);
//dscr
	}
	else {
		struct hdr_mac802_11 *dh;
		dh = HDR_MAC802_11(pktTx_);
		dh->dh_fc.fc_retry = 1;

		sendRTS(ETHER_ADDR(mh->dh_da));
		//printf("(%d)retxing data:%x..sendRTS..\n",index_,pktTx_);
//dscr
		//inc_cw();
		inc_scw();
		//printf("(%u) In RetransmitData, mhBackoff2 is started, fcw_ & scw_ is doubled: %u\n", index_, scw_);
                if(mhBackoff2_.busy() == 0) 
			mhBackoff2_.start(scw_, is_idle());
//dscr   
	}
}

/* ======================================================================
   Incoming Packet Routines
   ====================================================================== */
void
Mac802_11::send(Packet *p, Handler *h)
{
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);

	/* 
	 * drop the packet if the node is in sleep mode
	 XXX sleep mode can't stop node from sending packets
	 */
	EnergyModel *em = netif_->node()->energy_model();
	if (em && em->sleep()) {
		em->set_node_sleep(0);
		em->set_node_state(EnergyModel::INROUTE);
	}
	
	callback_ = h;
	sendDATA(p);
	sendRTS(ETHER_ADDR(dh->dh_da));

	/*
	 * Assign the data packet a sequence number.
	 */
	dh->dh_scontrol = sta_seqno_++;

	/*
	 *  If the medium is IDLE, we must wait for a DIFS
	 *  Space before transmitting.
	 */
//dscr
	if((mhBackoff_.busy() == 0) && (mhBackoff2_.busy() == 0)) {
		if(is_idle()) {     
			/*
			 * If we are already deferring, there is no
			 * need to reset the Defer timer.
			 */
			if(mhDefer_.busy() == 0){ 
 		                //printf("(%d)start defer timer at send()\n", index_);
//dscr
				phase = 2;
				threshTime = 0.0;  	//nonlinear decrement
//dscr
				mhDefer_.start(difs_);
			}
		}
		/*
		 * If the medium is NOT IDLE, then we start
		 * the backoff timer.
		 */
                else{
		  //printf("(%u) start backoff timer at position b1\n", index_);
                  mhBackoff_.start(fcw_, is_idle());
		}
	 } 
}

void
Mac802_11::recv(Packet *p, Handler *h)
{
	struct hdr_cmn *hdr = HDR_CMN(p);
//dscr
        struct hdr_mac802_11 *mh = HDR_MAC802_11(p);
        u_int8_t type = mh->dh_fc.fc_type;   
        u_int8_t subtype = mh->dh_fc.fc_subtype;
//dscr

	/*
	 * Sanity Check
	 */
	assert(initialized());

	/*
	 *  Handle outgoing packets.
	 */
	if(hdr->direction() == hdr_cmn::DOWN) {
                recvTime = Scheduler::instance().clock();
        
                //printf("\n (%u) at (%10.8g)--received a DOWN packet\n", index_, recvTime);

                send(p, h);
                return;
        }
	/*
	 *  Handle incoming packets.
	 *
	 *  We just received the 1st bit of a packet on the network
	 *  interface.
	 *
	 */

	/*
	 *  If the interface is currently in transmit mode, then
	 *  it probably won't even see this packet.  However, the
	 *  "air" around me is BUSY so I need to let the packet
	 *  proceed.  Just set the error flag in the common header
	 *  to that the packet gets thrown away.
	 */
         
	if(tx_active_ && hdr->error() == 0) {
		hdr->error() = 1;
	}

	if(rx_state_ == MAC_IDLE) {
	        //printf("\n (%u)--received a UP packet\n", index_);

		SET_RX_STATE(MAC_RECV);
		pktRx_ = p;

		/*
		 * Schedule the reception of this packet, in
		 * txtime seconds.
		 */
		mhRecv_.start(TX_Time(p));
	} else {
		/*
		 *  If the power of the incoming packet is smaller than the
		 *  power of the packet currently being received by at least
                 *  the capture threshold, then we ignore the new packet.
		 */
		if(pktRx_->txinfo_.RxPr / p->txinfo_.RxPr >= p->txinfo_.CPThresh) {
			capture(p);
		} else {
			collision(p);
		}
	}
}

void
Mac802_11::recv_timer()
{
	u_int32_t src; 
	hdr_cmn *ch = HDR_CMN(pktRx_);
	hdr_mac802_11 *mh = HDR_MAC802_11(pktRx_);
	u_int32_t dst = ETHER_ADDR(mh->dh_da);
	// XXX debug
	//struct cts_frame *cf = (struct cts_frame*)pktRx_->access(hdr_mac::offset_);
	//u_int32_t src = ETHER_ADDR(mh->dh_sa);
	
	u_int8_t  type = mh->dh_fc.fc_type;
	u_int8_t  subtype = mh->dh_fc.fc_subtype;
	double now;
        now = Scheduler::instance().clock();
 

	assert(pktRx_);
	assert(rx_state_ == MAC_RECV || rx_state_ == MAC_COLL);
	
	//fprintf(stderr, "(%d) Coming into recv_timer \n", index_);
        /*
         *  If the interface is in TRANSMIT mode when this packet
         *  "arrives", then I would never have seen it and should
         *  do a silent discard without adjusting the NAV.
         */
        if(tx_active_) {
                Packet::free(pktRx_);
                goto done;
        }

	/*
	 * Handle collisions.
	 */
	if(rx_state_ == MAC_COLL) {
		discard(pktRx_, DROP_MAC_COLLISION);
		//printf("(%d)set_nav at collision\n", index_);
		set_nav(usec(eifs_));
		goto done;
	}

	/*
	 * Check to see if this packet was received with enough
	 * bit errors that the current level of FEC still could not
	 * fix all of the problems - ie; after FEC, the checksum still
	 * failed.
	 */
	if( ch->error() != 0 ) {
		Packet::free(pktRx_);
                //printf("(%d)set_nav when channel error\n", index_);
		if(ch->error() == 1){
		    set_nav(usec(eifs_));
		}
		goto done;
	}

	/*
	 * IEEE 802.11 specs, section 9.2.5.6
	 *	- update the NAV (Network Allocation Vector)
	 */
	if(dst != (u_int32_t)index_) {
           //printf("(%d)set_nav for overhearing\n", index_);
	   set_nav(mh->dh_duration);
	   nav_category = 1;    //dscr
	   if(phase==2) {
	      if(subtype != MAC_Subtype_ACK){
 	        phase = 1;
		threshTime = 0.0;   //nonlinear decrement
		inc_cw();
		//printf("(%u) In recv_timer, fcw_ is doubled: %u\n", index_, fcw_);

                //printf("(%u) In recv_timer, overhear packet, phase changes from 2 to 1\n", index_);
                        					
		if(mhBackoff2_.busy()) mhBackoff2_.stop();
		if(mhDefer_.busy()) {
			//printf("(%u) stop DeferTimer in recv_timer() \n", index_);
			mhDefer_.stop();
		}
		//if(Pipelined_ == 2) {
		//	Pipelined_ = 0;
        	//	Tcl::instance().evalf("colreset");
		//	PWRand = PCWRSlot;			
		//}

		if((mhBackoff_.busy() == 0) && (pktRTS_ || pktTx_)){
	                //printf("(%u) Start backoff timer\n", index_);
       			mhBackoff_.start(fcw_, is_idle());
		} else{ 
		        CHECK_BACKOFF_TIMER();      
                }
	    }
	   }
	}
//dscr
        /* tap out - */
        if (tap_ && type == MAC_Type_Data &&
            MAC_Subtype_Data == subtype ) 
		tap_->tap(pktRx_);
	/*
	 * Adaptive Fidelity Algorithm Support - neighborhood infomation 
	 * collection
	 *
	 * Hacking: Before filter the packet, log the neighbor node
	 * I can hear the packet, the src is my neighbor
	 */
	if (netif_->node()->energy_model() && 
	    netif_->node()->energy_model()->adaptivefidelity()) {
		src = ETHER_ADDR(mh->dh_sa);
		netif_->node()->energy_model()->add_neighbor(src);
	}
	/*
	 * Address Filtering
	 */
	if(dst != (u_int32_t)index_ && dst != MAC_BROADCAST) {
		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 */
		discard(pktRx_, "---");
		goto done;
	}

	switch(type) {

	case MAC_Type_Management:
		discard(pktRx_, DROP_MAC_PACKET_ERROR);
		goto done;
		break;

	case MAC_Type_Control:
		switch(subtype) {
		case MAC_Subtype_RTS:
                        //printf("\n(%u) received RTS!\n", index_);
			recvRTS(pktRx_);
			break;
		case MAC_Subtype_CTS:
                        //printf("\n(%u) received CTS!\n", index_);
			recvCTS(pktRx_);
			break;
		case MAC_Subtype_ACK:
                        //printf("\n(%u) received ACK at %10.8g!\n", index_, now);
			recvACK(pktRx_);
			break;
		default:
			fprintf(stderr,"recvTimer1:Invalid MAC Control Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(subtype) {
		case MAC_Subtype_Data:
                        //printf("\n(%u) received Data!\n", index_);
			recvDATA(pktRx_);
			break;
		default:
			fprintf(stderr, "recv_timer2:Invalid MAC Data Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "recv_timer3:Invalid MAC Type %x\n", subtype);
		exit(1);
	}
 done:
	pktRx_ = 0;
	rx_resume();
}


void
Mac802_11::recvRTS(Packet *p)
{
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);

	if(tx_state_ != MAC_IDLE) {
		discard(p, DROP_MAC_BUSY);
		return;
	}

	/*
	 *  If I'm responding to someone else, discard this RTS.
	 */
	if(pktCTRL_) {
		discard(p, DROP_MAC_BUSY);
		return;
	}

	sendCTS(ETHER_ADDR(rf->rf_ta), rf->rf_duration);

	/*
	 *  Stop deferring - will be reset in tx_resume().
	 */
	if(mhDefer_.busy()) {
	    //printf("(%u) Stop DeferTimer at recvRTS() \n", index_);
	    mhDefer_.stop();
	}
//dscr 
	if(mhBackoff2_.busy()) {
		//printf("(%u) mhBackoff2 is stopped at recvRTS()\n", index_);
		mhBackoff2_.stop();
	}
	if(phase == 2) {
		phase = 1;
		threshTime = 0.0;   //nonlinear decrement
		//printf("(%u) As receiver, phase changes from 2 to 1 at recvRTS()\n", index_);
                inc_cw();
                //printf("(%u) In recvRTS, fcw_ is doubled: %u\n", index_, fcw_);
	}
	recv_ = 1;
	if(mhBackoff_.busy() && !mhBackoff_.paused())
	   mhBackoff_.pause();
//dscr

	tx_resume();

	mac_log(p);
}


void
Mac802_11::recvCTS(Packet *p)
{
	if(tx_state_ != MAC_RTS) {
		discard(p, DROP_MAC_INVALID_STATE);
		return;
	}

	assert(pktRTS_);
	Packet::free(pktRTS_); pktRTS_ = 0;

	assert(pktTx_);
	// debug
	//struct hdr_mac802_11 *mh = HDR_MAC802_11(pktTx_);
	//printf("(%d):recvCTS:pktTx_-%x,mac-subtype-%d & pktCTS_:%x\n",index_,pktTx_,mh->dh_fc.fc_subtype,p);
	
	mhSend_.stop();

        cwtime = Scheduler::instance().clock();
                
        //printf("%d %10.9g 5 %d\n", index_, cwtime, scw_);
        //printf("%d %10.9g 7 %d\n", index_, cwtime, rand);
 
	/*
	 * The successful reception of this CTS packet implies
	 * that our RTS was successful.  Hence, we can reset
	 * the Short Retry Count and the CW.
	 */
	ssrc_ = 0;
	//rst_cw();
//dscr
//	rst_scw();

        cwtime = Scheduler::instance().clock();
                 
        //printf("%d %10.9g 2 %d\n", index_,cwtime, scw_);
//dscr
	tx_resume();

	mac_log(p);
}

void
Mac802_11::recvDATA(Packet *p)
{
	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	u_int32_t dst, src, size;

	{	struct hdr_cmn *ch = HDR_CMN(p);

		dst = ETHER_ADDR(dh->dh_da);
		src = ETHER_ADDR(dh->dh_sa);
		size = ch->size();

		/*
		 * Adjust the MAC packet size - ie; strip
		 * off the mac header
		 */
		ch->size() -= ETHER_HDR_LEN;
		ch->num_forwards() += 1;
	}
	/*
	 *  If we sent a CTS, clean up...
	 */
	if(dst != MAC_BROADCAST) {
//dscr
        	if(mhBackoff2_.busy()) {
                	//printf("(%u) mhBackoff2 is stopped at recvDATA()\n", index_);
                	mhBackoff2_.stop();
        	}
        	if(phase == 2) {
                	phase = 1;
			threshTime = 0.0;   //nonlinear decrement 
                	//printf("(%u) As receiver, phase changes from 2 to 1 at recvDATA()\n", index_);
                	inc_cw();
                	//printf("(%u) In recvDATA, fcw_ is doubled: %u\n", index_, fcw_);
        	}
        	recv_ = 1;
        	if(mhBackoff_.busy() && !mhBackoff_.paused())
           		mhBackoff_.pause();
//dscr
		if(size >= macmib_->RTSThreshold) {
			if (tx_state_ == MAC_CTS) {
				assert(pktCTRL_);
				Packet::free(pktCTRL_); pktCTRL_ = 0;
				mhSend_.stop();
				/*
				 * Our CTS got through.
				 */
				//printf("(%d): RECVING DATA!\n",index_);
				ssrc_ = 0;
//dscr
				rst_scw();
//dscr
		                cwtime = Scheduler::instance().clock();
                 
                		//printf("%d %10.9g 1 %d\n", index_,cwtime, fcw_);

			}
			else {
				discard(p, DROP_MAC_BUSY);
				//printf("(%d)..discard DATA\n",index_);
				return;
			}
			sendACK(src);
			tx_resume();
//dscr ???????????????????????????????????????????????
//                        if(phase == 2)
//				phase = 1;
//dscr ???????????????????????????????????????????????

		}
		/*
		 *  We did not send a CTS and there's no
		 *  room to buffer an ACK.
		 */
		else {
			if(pktCTRL_) {
				discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);
			if(mhSend_.busy() == 0)
				tx_resume();
		}
	}

	/* ============================================================
	    Make/update an entry in our sequence number cache.
	   ============================================================ */

	/* Changed by Debojyoti Dutta. This upper loop of if{}else was 
	   suggested by Joerg Diederich <dieder@ibr.cs.tu-bs.de>. 
	   Changed on 19th Oct'2000 */

        if(dst != MAC_BROADCAST) {
                if (src < (u_int32_t) cache_node_count_) {
                        Host *h = &cache_[src];

                        if(h->seqno && h->seqno == dh->dh_scontrol) {
                                discard(p, DROP_MAC_DUPLICATE);
                                return;
                        }
                        h->seqno = dh->dh_scontrol;
                } else {
			static int count = 0;
			if (++count <= 10) {
				printf ("MAC_802_11: accessing MAC cache_ array out of range (src %u, dst %u, size %d)!\n", src, dst, cache_node_count_);
				if (count == 10)
					printf ("[suppressing additional MAC cache_ warnings]\n");
			};
		};
	}

	/*
	 *  Pass the packet up to the link-layer.
	 *  XXX - we could schedule an event to account
	 *  for this processing delay.
	 */
	p->incoming = 1;
	uptarget_->recv(p, (Handler*) 0);
}


void
Mac802_11::recvACK(Packet *p)
{
//	double threshTime;	
	struct hdr_cmn *ch = HDR_CMN(p);

	if(tx_state_ != MAC_SEND) {
	discard(p, DROP_MAC_INVALID_STATE);
	return;
	}
	//printf("(%d)...................recving ACK:%x\n",index_,p);

#ifdef YX_DELAY
//dscr
        ackTime = Scheduler::instance().clock();
        macDelay = ackTime - recvTime; 
	if (ackTime > StartTime_) printf("%10.8g  %10.8g\n", ackTime - StartTime_, macDelay);
//dscr
#endif
	
	assert(pktTx_);
	Packet::free(pktTx_); pktTx_ = 0;

	mhSend_.stop();

	/*
	 * The successful reception of this ACK packet implies
	 * that our DATA transmission was successful.  Hence,
	 * we can reset the Short/Long Retry Count and the CW.
	 */
	if((u_int32_t) ch->size() <= macmib_->RTSThreshold)
		ssrc_ = 0;
	else
		slrc_ = 0;

	/*
	 * Backoff before sending again.
	 */
	rst_cw();

//dscr
	phase = 1;
	threshTime = 0.0;
        Tcl::instance().evalf("colreset");
	rst_scw();
        cwtime = Scheduler::instance().clock();
                 
        //printf("%d receiving ACK at time %10.9g, fcw_ is: %u, scw_ is: %u\n", index_, cwtime, fcw_, scw_);

	assert(mhBackoff_.busy() == 0);
        //printf("(%u) After RecvACK, phase is 1, start backoff timer\n", index_);

	mhBackoff_.start(fcw_, is_idle());
/*dscr --------this will let the sender who just tranmitted enter 2nd phase immediately,
---------------However, it cause some unfairness ----------------------------------
        mhBackoff_.start(fcw_, 0);
        threshTime = K * phymib_->SlotTime;
        if (FR_Time > threshTime) {
            FR_Time -= threshTime;
        } else {
            FR_Time = 0.0;
        }
	mhBackoff_.resume(FR_Time);
dscr */		
	tx_resume();
	
	mac_log(p);
}

void
Mac802_11::Backoff_Remain(double rtime)
{
	FR_Time = rtime;
}


